/****************************************************************************
**  LinkTest                                                               **
*****************************************************************************
**  Copyright (c) 2008-2022                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
#include "config.h"
#include "compiler.h"
#include "vcluster.h"
#include "vcluster_helper.h"
#if HAVE_VCLUSTER_TCP == 1
#include "vcluster_tcp.h"
#endif
#if HAVE_VCLUSTER_MPI == 1
#include "vcluster_mpi.h"
#endif
#if HAVE_VCLUSTER_IBVERBS == 1
#include "vcluster_ibverbs.h"
#endif
#if HAVE_VCLUSTER_PSM2 == 1
#include "vcluster_psm2.h"
#endif
#if HAVE_VCLUSTER_UCP == 1
#include "vcluster_ucp.h"
#endif
#if HAVE_VCLUSTER_PORTALS == 1
#include "vcluster_portals.h"
#endif
#if HAVE_VCLUSTER_CUDA == 1
#include "vcluster_cuda.h"
#endif
#include "error.h"
#include "environ.h"
#include "cmdline.h"
#include "timing.h"
#include "output_sion.h"
#include "stats.h"
#include <cstring>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include "utils.h"


VirtualCluster::VirtualCluster(const std::string name)
: name_(name), nameSize_(name_.size()+1), hostname_(determineHostname()), hostnameSize_(hostname_.size()+1), cpuID_(determineCPUID())
{}

VirtualCluster::~VirtualCluster()
{}

std::string VirtualCluster::name(){
    return name_;
}

const std::string& VirtualCluster::nameRef(){
    return name_;
}

int32_t VirtualCluster::nameSize(){
    return nameSize_;
}

const int32_t& VirtualCluster::nameSizeRef(){
    return nameSize_;
}

std::string VirtualCluster::hostname(){
    return hostname_;
}

const std::string& VirtualCluster::hostnameRef(){
    return hostname_;
}

int32_t VirtualCluster::hostnameSize(){
    return hostnameSize_;
}

const int32_t& VirtualCluster::hostnameSizeRef(){
    return hostnameSize_;
}

int32_t VirtualCluster::cpuID(){
    return cpuID_;
}

const int32_t& VirtualCluster::cpuIDRef(){
    return cpuID_;
}

int VirtualCluster::numHostRanks(){
    return numHostRanks_;
}

int VirtualCluster::numLocalRanks(){
    return numLocalRanks_;
}

int VirtualCluster::hostRank(){
    return hostRank_;
}

int VirtualCluster::localRank(){
    return localRank_;
}

std::shared_ptr<int[]> VirtualCluster::hostLocalRanks(){
    return hostLocalRanks_;
}

int VirtualCluster::barrier(){
    return vcluster_helper_barrier(this);
}

int VirtualCluster::bcast(int root, MemoryBuffer& buf){
    return vcluster_helper_bcast(this, root, buf);
}

int VirtualCluster::gather(int root, MemoryBuffer& dst, MemoryBuffer& src){
    return vcluster_helper_gather(this, root, dst, src);
}

int VirtualCluster::scatter(int root, MemoryBuffer& dst, MemoryBuffer& src){
    return vcluster_helper_scatter(this, root, dst, src);
}

int VirtualCluster::reduce(int root, MemoryBuffer& dst, MemoryBuffer& src, int op){
    return vcluster_helper_reduce(this, root, dst, src, op);
}

int VirtualCluster::kpingpong(int from, int to, MemoryBuffer& buf, int num_msg, double* timing){
    double tv = 0;

    if(rank() == from) {
        tv = walltime();

        for (auto i = 0; i < num_msg; ++i) {
            EXEC_IFFAIL(send(to, buf), fatal("send failed."));

            EXEC_IFFAIL(recv(to, buf), fatal("recv failed."));
        }

        tv = walltime() - tv;
        tv = tv / (2.0*num_msg);
    } else if(rank() == to) {
        for (auto i = 0; i < num_msg; ++i) {
            EXEC_IFFAIL(recv(from, buf), fatal("recv failed."));

            EXEC_IFFAIL(send(from, buf), fatal("send failed."));
        }
    } else {
        error("Invalid rank.");
        return ERROR;
    }

    if(timing)
        *timing = tv;

    return SUCCESS;
}

int VirtualCluster::kUniDir(const int from, const int to,
                            MemoryBuffer& buf1, MemoryBuffer& buf2,
                            const int num_msg, double* const timing,
                            const bool doBarrier
                           ){
    /* Unidirectional ping test in which in which the origin rank times
       the time needed to send N messages back-to-back to its partner,
       who returns a receipt message upon receiving all messages, till
       the origin rank receives said receipt. The returned time is
       averaged over the number of messages sent from the origin rank to
       its partner.
     */
    double tv=0.0;
    if(doBarrier)barrier();
    if(rank()==from){
        tv=walltime();
        for(int i=0;i<num_msg;i++){
            auto ret=send(to,buf1);
            if(unlikely(ret)){fatal("Send failed.");return ERROR;}
        }
        auto ret=recv(to,buf2);
        if(unlikely(ret)){fatal("Receive failed.");return ERROR;}
        tv=(walltime()-tv)/num_msg;
    }else if(rank()==to){
        for(int i=0;i<num_msg;i++){
            auto ret=recv(from,buf1);
            if(unlikely(ret)){fatal("Receive failed.");return ERROR;}
        }
        auto ret=send(from,buf2);
        if(unlikely(ret)){fatal("Send failed.");return ERROR;}
    }else{
        error("Invalid rank.");
        return ERROR;
    }
    if(timing!=NULL) *timing = tv;
    return SUCCESS;
}
int VirtualCluster::kUniDirMultiBuf(const int from,const int to,
                                    MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                    const int num_msg, double* const timing,
                                    const bool doBarrier
                                   ){
    fatal("not yet implemented.");
    return ERROR;
}
int VirtualCluster::kUniDirLimitedMultiBuf(const int from,const int to,
                                           MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                           const int num_msg, double* const timing,
                                           const bool doBarrier
                                          ){
    fatal("not yet implemented.");
    return ERROR;
}

uint64_t VirtualCluster::kpingpong_minimal_buffer_overhead() const
{
    return 0;
}

int VirtualCluster::kalltoall(MemoryBuffer& buf, int len_msg, int num_msg, double *timing)
{
    fatal("not yet implemented.");
    return ERROR;
}

int VirtualCluster::kbipingpong(int from, int to,
                                MemoryBuffer& buf1,
                                MemoryBuffer& buf2,
                                int num_msg, double* timing)
{
    fatal("not yet implemented.");
    return ERROR;
}

int VirtualCluster::linktest_kpingpong(const int from, const int to,
                                       MemoryBuffer& buf,
                                       const struct linktest_args* const args,
                                       double* const time){
    int err;
    debug("VirtualCluster::linktest_kpingpong from %d to %d", from, to);
    // Warmup
    if(args->num_warmup_msg!=0){ //No use doing this if there are no warm-up messages.
        double dummy;
        err=kpingpong(from,to,buf,
                      args->num_warmup_msg,
                      &dummy);
        if(unlikely(err))return ERROR;
    }
    err=kpingpong(from,to,buf,args->num_msg,time);
    if(unlikely(err))return err;
    if(args->check_buffers){
        if(unlikely(buf.check())){
            int other=(rank()==from)?to:from;
            warn("Rank %d encountered malformed buffer after ping pong with rank %d.",rank(),other);
            buf.fill();
        }
    }
    return SUCCESS;
}

int VirtualCluster::linktest_kUniDir(const int from, const int to,
                                     MemoryBuffer& buf1, MemoryBuffer& buf2,
                                     const struct linktest_args* const args,
                                     double* const time, const int doBarrier
                                    ){
    int err;
    // Warmup
    if(args->num_warmup_msg!=0){ //No use doing this if there are no warm-up messages.
        double dummy;
        err=kUniDir(from,to,buf1,buf2,args->num_warmup_msg,&dummy,doBarrier);
        if(unlikely(err))return err;
    }

    err=kUniDir(from,to,buf1,buf2,args->num_msg,time,doBarrier);
    if(unlikely(err))return err;
    if(args->check_buffers){
        if(rank()==to){
            if(unlikely(buf1.check())){
                warn("Rank %d received malformed buffer during last unidirectional ping-pong with rank %d.",rank(),from);
                buf1.fill();
            }
        }else if(rank()==from){
            if(unlikely(buf2.check())){
                warn("Rank %d encountered malformed receipt after unidirectional ping-pong with rank %d.",rank(),to);
                buf2.fill();
            }
        }
    }
    return SUCCESS;
}
int VirtualCluster::linktest_kUniDirMultiBuf(const int from, const int to,
                                             MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                             const struct linktest_args* const args,
                                             double* const time, const int doBarrier
                                            ){
    int err;
    // Warmup
    if(args->num_warmup_msg!=0){ //No use doing this if there are no warm-up messages.
        double dummy;
        auto err=kUniDirMultiBuf(from,to,buf_multi,buf2,args->num_warmup_msg,&dummy,doBarrier);
        if (unlikely(err))return ERROR;
    }

    err=kUniDirMultiBuf(from,to,buf_multi,buf2,args->num_msg,time,doBarrier);
    if(unlikely(err))return err;
    if(args->check_buffers){
        if(rank()==to){
            std::size_t buffer, byte;
            if(unlikely(buf_multi.check(&buffer,&byte))){
                warn("Rank %d received malformed buffer %zu (byte %zu) during last unidirectional ping-pong with rank %d.",rank(),buffer,byte,from);
                buf_multi.fill();
            }
        }else if(rank()==from){
            if(unlikely(buf2.check())){
                warn("Rank %d encountered malformed receipt after unidirectional ping-pong with rank %d.",rank(),to);
                buf2.fill();
            }
        }
    }
    return SUCCESS;
}
int VirtualCluster::linktest_kUniDirLimitedMultiBuf(const int from, const int to,
                                                    MemoryBufferMulti& buf_multi, MemoryBuffer& buf2,
                                                    const struct linktest_args* const args,
                                                    double* const time, const int doBarrier
                                                   ){
    int err;
    // Warmup
    if(args->num_warmup_msg!=0){ //No use doing this if there are no warm-up messages.
        double dummy;
        err=kUniDirLimitedMultiBuf(from,to,buf_multi,buf2,args->num_warmup_msg,&dummy,doBarrier);
        if (unlikely(err))return ERROR;
    }

    err=kUniDirLimitedMultiBuf(from,to,buf_multi,buf2,args->num_msg,time,doBarrier);
    if(unlikely(err))return err;
    if(args->check_buffers){
        if(rank()==to){
            std::size_t buffer, byte;
            if(unlikely(buf_multi.check(&buffer,&byte))){
                warn("Rank %d received malformed buffer %zu (byte %zu) during last unidirectional ping-pong with rank %d.",rank(),buffer,byte,from);
                buf_multi.fill();
            }
        }else if(rank()==from){
            if(unlikely(buf2.check())){
                warn("Rank %d encountered malformed receipt after unidirectional ping-pong with rank %d.",rank(),to);
                buf2.fill();
            }
        }
    }
    return SUCCESS;
}

int VirtualCluster::linktest_kalltoall(MemoryBuffer& buf,
                                       const struct linktest_args* const args,
                                       double* const time){
    /* Warmup */
    EXEC_NOFAIL(kalltoall(buf, args->len_msg, args->num_warmup_msg, NULL));

    /* Perform Actual Test */
    //TODO: Check Buffers
    return kalltoall(buf, args->len_msg, args->num_msg, time);
}

int VirtualCluster::linktest_kbipingpong(const int from, const int to,
                                         MemoryBuffer& buf1, MemoryBuffer& buf2,
                                         const struct linktest_args* const args,
                                         double* const time){
    int err;
    // Warmup
    if(args->num_warmup_msg!=0){ //No use doing this if there are no warm-up messages.
        double dummy;
        err=kbipingpong(from,to,buf1,buf2,args->num_warmup_msg,&dummy);
        if (unlikely(err)) return err;
    }

    err=kbipingpong(from,to,buf1,buf2,args->num_msg,time);
    if(unlikely(err))return err;
    if(args->check_buffers){
        if(unlikely(buf1.check())){
            int other=(rank()==from)?to:from;
            warn("Rank %d encountered malformed first buffer after bidirectional ping-pong with rank %d.",rank(),other);
            buf1.fill();
        }
        if(unlikely(buf2.check())){
            int other=(rank()==from)?to:from;
            warn("Rank %d encountered malformed second buffer after bidirectional ping-pong with rank %d.",rank(),other);
            buf2.fill();
        }
    }
    return SUCCESS;
}

const std::vector<std::string> VirtualCluster::impls = {
    #if 1 == HAVE_VCLUSTER_TCP
    "tcp",
    #endif
    #if 1 == HAVE_VCLUSTER_MPI
    "mpi",
    #endif
    #if 1 == HAVE_VCLUSTER_IBVERBS
    "ibverbs",
    #endif
    #if 1 == HAVE_VCLUSTER_PSM2
    "psm2",
    #endif
    #if 1 == HAVE_VCLUSTER_UCP
    "ucp",
    #endif
    #if 1 == HAVE_VCLUSTER_PORTALS
    VirtualClusterPortals::NAME,
    #endif
    #if 1 == HAVE_VCLUSTER_CUDA
    "cuda",
    #endif
};

VirtualCluster*  VirtualCluster::factory(const std::string& name){
#if 1 == HAVE_VCLUSTER_TCP
    if ("tcp" == name) {
        return new VirtualClusterTCP(name);
    } else 
#endif
#if 1 == HAVE_VCLUSTER_MPI
    if ("mpi" == name) {
        return new VirtualClusterMPI(name);
    } else 
#endif
#if 1 == HAVE_VCLUSTER_IBVERBS
    if ("ibverbs" == name) {
        return new VirtualClusterIbverbs(name);
    } else 
#endif
#if 1 == HAVE_VCLUSTER_PSM2
    if ("psm2" == name) {
        return new VirtualClusterPSM2(name);
    } else
#endif
#if 1 == HAVE_VCLUSTER_UCP
    if ("ucp" == name) {
        return new VirtualClusterUCP(name);
    } else
#endif
#if 1 == HAVE_VCLUSTER_PORTALS
    if (VirtualClusterPortals::NAME == name) {
        return new VirtualClusterPortals();
    } else
#endif
#if 1 == HAVE_VCLUSTER_CUDA
    if ("cuda" == name) {
        return new VirtualClusterCUDA(name);
    } else
#endif
    {}

    debug("Unknown VirtualCluster name \"%s\"", name.c_str());
    return nullptr;
}


VirtualClusterWithHelper::VirtualClusterWithHelper(const std::string& name)
: VirtualCluster(name), helper_(nullptr)
{
}

void VirtualClusterWithHelper::set_helper_pointer(VirtualCluster* helper)
{
    helper_ = helper;
}

int VirtualClusterWithHelper::rank()
{
    if (unlikely(!helper_))
        fatal("rank() called on a VirtualClusterWithHelper with undefined helper");
    return helper_->rank();
}

int VirtualClusterWithHelper::size()
{
    if (unlikely(!helper_))
        fatal("size() called on a VirtualClusterWithHelper with undefined helper");
    return helper_->size();
}

int VirtualClusterWithHelper::send(int dst, MemoryBuffer& buf)
{
    if (unlikely(!helper_))
        fatal("send() called on a VirtualClusterWithHelper with undefined helper");
    return helper_->send(dst, buf);
}

int VirtualClusterWithHelper::recv(int src, MemoryBuffer& buf)
{
    if (unlikely(!helper_))
        fatal("recv() called on a VirtualClusterWithHelper with undefined helper");
    return helper_->recv(src, buf);
}

const std::string& VirtualCluster::get_vcluster_impl_name(char** argv, const std::string& name)
{
    std::string requestedImpl;

    // check 'name'
    if(name != "") {
        requestedImpl = name;
    }

    // check executable extension
    std::string executableName(argv[0]);
    const std::string dot{"."};
    auto pos = executableName.find(dot);
    if(pos != std::string::npos) {
        requestedImpl = executableName.substr(pos);
    }

    // Check environment
    const char* envName;
    if(read_environ_str(LINKTEST_ENVIRON_PREFIX "VCLUSTER_IMPL", &envName) == SUCCESS) {
        requestedImpl = {envName};
    }

    #if 1 == HAVE_VCLUSTER_TCP
    debug("HAVE_VCLUSTER_TCP == 1");
    #endif
    #if 1 == HAVE_VCLUSTER_MPI
    debug("HAVE_VCLUSTER_MPI == 1");
    #endif
    #if 1 == HAVE_VCLUSTER_IBVERBS
    debug("HAVE_VCLUSTER_IBVERBS == 1");
    #endif
    #if 1 == HAVE_VCLUSTER_PSM2
    debug("HAVE_VCLUSTER_PSM2 == 1");
    #endif
    #if 1 == HAVE_VCLUSTER_UCP
    debug("HAVE_VCLUSTER_UCP == 1");
    #endif
    #if 1 == HAVE_VCLUSTER_PORTALS
    debug("HAVE_VCLUSTER_PORTALS == 1");
    #endif
    #if 1 == HAVE_VCLUSTER_CUDA
    debug("HAVE_VCLUSTER_CUDA == 1");
    #endif
    debug("requestedImpl = %s", requestedImpl.c_str());

    for(const auto& impl : VirtualCluster::impls) {
        if(impl == requestedImpl) {
            return impl;
        }
    }
    
    error("Internal Error: Unexpected negative executable-name-suffix offset encountered.");
    return VirtualCluster::impls.at(0); // unreachable code to supress warning
}

int VirtualCluster::write_parallel(const linktest_args* args, const std::vector<LinktestStats>& statsVec)
{
    return linktest_output_sion_parallel(this, args, statsVec);
}

int VirtualCluster::write_funnelled(const linktest_args* args, const std::vector<LinktestStats>& statsVec)
{
    return linktest_output_sion_funnelled(this, args, statsVec);
}

void VirtualCluster::getHostAndLocalRank(){
    std::int64_t buf64x4[4];
    std::int32_t tmp32=hostnameSize();
    /***************************************************/
    /* Determine unique hostnames and associated ranks */
    /***************************************************/
    if(rank()==0){
        /********************/
        /* Gather Hostnames */
        /********************/
        std::vector<std::int32_t> hostnameLengths;
        std::vector<std::string> hostnames;
        hostnameLengths.resize(size());
        gather(0, hostnameLengths.data(), &tmp32, 1); //Gather hostname lengths
        hostnames.resize(size());
        hostnames[0]=hostname();
        for(int rank = 1; rank < size(); rank++){
            std::vector<char> buf(hostnameLengths[rank]);
            recv(rank, buf.data(), hostnameLengths[rank]);
            hostnames[rank] = std::string(buf.begin(), buf.end());
        }

        /***************************************************/
        /* Determine unique hostnames and associated ranks */
        /***************************************************/
        std::vector<std::string> uhostnames;
        uhostnames.push_back(hostnames[0]);
        std::vector<std::vector<int>> ranks(1);
        ranks[0].push_back(0);
        for(int i=1;i<size();i++){
            int j=uhostnames.size()-1;
            for(;j>=0;j--){ //Iterate backwards as hostnames are likely to repeat
                if(hostnames[i]!=uhostnames[j]) continue;
                else{
                    ranks[j].push_back(i);
                    break;
                }
            }
            if(j<0){
                uhostnames.push_back(hostnames[i]);
                ranks.resize(uhostnames.size());
                ranks[uhostnames.size()-1].push_back(i);
            }
        }
#ifdef DEBUG_Combine_Hosts
        std::stringstream buf;
        for(std::vector<std::string>::size_type i=0;i<uhostnames.size();i++){
            buf << uhostnames[i].c_str() << ": " << ranks[i];
            debug("[%3d] Ranks on %s", 0, buf.str().c_str());buf.str(std::string());
        }
#endif
        /******************************************************/
        /* Check that all hosts have the same number of ranks */
        /******************************************************/
        if(ranks.size()%2!=0) fatal("An even number of hosts is required!");
        for(std::vector<std::vector<int>>::size_type i=0;i<ranks.size();i++) if(ranks[0].size()!=ranks[i].size()) fatal("Hosts have differing amounts of ranks!");

        /**************************************/
        /* Broadcast information to all ranks */
        /**************************************/
        // Buffer for communication
        buf64x4[0]=(std::int64_t) ranks.size();
        buf64x4[1]=0;
        buf64x4[2]=(std::int64_t) ranks[0].size();
        hostLocalRanks_ = std::make_unique<int[]>(buf64x4[0]*buf64x4[2]);
        for(std::int64_t i=0;i<buf64x4[0];i++)for(std::int64_t j=0;j<buf64x4[2];j++) hostLocalRanks_[i*buf64x4[2]+j]=ranks[i][j];
        // Set for rank 0
        numHostRanks_ =buf64x4[0];
        hostRank_     =buf64x4[1];
        numLocalRanks_=buf64x4[2];
        localRank_    =0;
#ifdef DEBUG_Combine_Hosts
        debug("[%3d] Host: %s Num. Host ranks: %d Host rank: %d Num. Local Ranks: %d Local Rank: %d",0,hostnames[0].c_str(),numHostRanks_,hostRank_,numLocalRanks_,localRank_);
#endif
        // Loop over other ranks on first host
        for(std::vector<int>::size_type j=1;j<ranks[0].size();j++){
            buf64x4[3]=(std::int64_t)j;
            send(ranks[0][j], buf64x4, 4);
#ifdef DEBUG_Combine_Hosts
            debug("[%3d] Host: %s Num. Host ranks: %d Host rank: %d Num. Local Ranks: %d Local Rank: %d",ranks[0][j],uhostnames[0].c_str(),buf64x4[0],buf64x4[1],buf64x4[2],buf64x4[3]);
#endif
        }
        // Loop over remaining ranks
        for(std::vector<std::vector<int>>::size_type i=1;i<ranks.size();i++){
            buf64x4[1]=(std::int64_t)i;
            buf64x4[2]=ranks[i].size();
            for(std::vector<int>::size_type j=0;j<ranks[i].size();j++){
                buf64x4[3]=(std::int64_t)j;
                send(ranks[i][j], buf64x4, 4);
#ifdef DEBUG_Combine_Hosts
                debug("[%3d] Host: %s Num. Host ranks: %d Host rank: %d Num. Local Ranks: %d Local Rank: %d",ranks[i][j],uhostnames[i].c_str(),buf64x4[0],buf64x4[1],buf64x4[2],buf64x4[3]);
#endif
            }
        }
        bcast(0, hostLocalRanks_.get(), buf64x4[0]*buf64x4[2]);
    }else{
        gather(0, &tmp32, &tmp32, 1); //Gather hostname lengths
        send(0, hostname().c_str(), hostnameSize());  //Send hostname
        recv(0, buf64x4, 4);
        numHostRanks_ =buf64x4[0];
        hostRank_     =buf64x4[1];
        numLocalRanks_=buf64x4[2];
        localRank_    =buf64x4[3];
        hostLocalRanks_ = std::make_unique<int[]>(buf64x4[0]*buf64x4[2]);
        bcast(0, hostLocalRanks_.get(), buf64x4[0]*buf64x4[2]);
    }
}
